#include <stdio.h>
#include <stdlib.h>
#include <string.h>

/*
 * Constants
 * ---------
 * Norm:           The maximal "infinity norm" of node labels in decompositions
 * MaxInputLength: The maximal length of the string to be tested
 * AlphabetSize:   The size of the input alphabet
 * FirstCharacter: The first chracter in the alphabet;
 *                 the characters are assumed to be consecutive
 * InputString:    The input string to be tested
 */

#define Norm 2
#define MaxInputLength 100
#define AlphabetSize 3
#define FirstCharacter 'a'

#define InputString "aaaaabbbbbbbbbbbbbbaaaaaaaaaaaaaaaaaaacccccccccccccccccccccccccccccbbbbbbbbbbbbbbbaaaaa"
                /* a^5 b^{14} a^{19} c^{29} b^{15} a^5 */

/*
 * Global variables
 * ----------------
 * derived: 4-dimensional array holding integers;
 * If w is the input string,
 * derived[i][j][k][l] == 1 means that a decomposition
 * within the allowed norm with root label (w[i,j], w[k,l])
 * has been found.
 * derived[i][j][k][l] == 0 means that it has been established
 * that there is no decomposition within the allowed norm with
 * (w[i,j], w[k,l]) as root label.
 */ 

int derived[MaxInputLength+1][MaxInputLength+1][MaxInputLength+1][MaxInputLength+1];

/*
 * Function prototypes
 */

void makeCharacterCountTable(char input[], int inputLength, int table[][MaxInputLength+1][MaxInputLength+1]);

int recognize(int table[][MaxInputLength+1][MaxInputLength+1], int n);

int derive(int table[][MaxInputLength+1][MaxInputLength+1], int i, int j, int k, int l);

int exceedsNorm(int counts[]);

int member(int n, int list[], int l);

int isInMIX(int table[][MaxInputLength+1][MaxInputLength+1], int inputLength);

/* Main program */

main()
{
  int i, j, k, l;
  int inputLength;

  int table[AlphabetSize][MaxInputLength+1][MaxInputLength+1];

  char input[] = InputString;
  
  inputLength = strlen(input);

  printf("%s\n\n", input);

  /*
    First convert the input string into a character count table representing
    how many times each character occurs within each substring of the input string.
  */
  makeCharacterCountTable(input, inputLength, table);

  if (isInMIX(table, inputLength)) {
    printf("Input is in MIX.\n\n");
  } else {
    printf("Input is not in MIX.\n\n");
    return;
  }

  if (recognize(table, inputLength) == 1) {
    printf("\nYes\n");
  } else {
    printf("\nNo\n");
  }

  return;
}

/*
 * Function: makeCharacterCountTable
 * ---------------------------------
 * For each character in the alphabet, records how many times
 * it occurs in the interval [i,j], for each i,j such that
 * 0 <= i <= j <= inputLength.
 */

void makeCharacterCountTable(char input[], int inputLength, int table[][MaxInputLength+1][MaxInputLength+1])
{
  int i, j, c;

  for (j = 0; j < inputLength + 1; j++) {
    for (c = 0; c < AlphabetSize; c++) {
      table[c][j][j] = 0;
      if (input[j-1] - FirstCharacter == c) {
	for (i = 0; i < j; i++) {
	  table[c][i][j] = table[c][i][j-1] + 1;
	}
      } else {
	for (i = 0; i < j; i++) {
	  table[c][i][j] = table[c][i][j-1];
	}
      }
    }
  }
}

/*
 * Function: recognize
 * -------------------
 * Given a character count table, determine
 * whether the string in the interval [0,inputLength] 
 * has a decomposition within the allowed norm.
 * It initializes the memo table derived[][], splits
 * the string into two, and call derive.
 */

int recognize(int table[][MaxInputLength+1][MaxInputLength+1], int inputLength)
{
  int i, j, k, l;

  /* initialize derived[][][][] */
  for (i = 0; i < inputLength+1 ; i++) {
    for (j = i; j < inputLength+1; j++) {
      for (k = j; k < inputLength+1; k++) {
	for (l = k; l < inputLength+1; l++) {
	  derived[i][j][k][l] = -1;
	}
      }
    }
  }

  for (i = 0; i <= inputLength; i++) {
    printf("Splitting input at %d\n", i);
    if (derive(table, 0, i, i, inputLength) == 1) return 1;
  }

  return 0;
}

/*
 * Function: derive
 * ----------------
 * Given a character count table, determine whether
 * the pair of strings in the intervals [i,j] and [k,l] has 
 * a decomposition within the allowed norm.  This function uses
 * the four-dimensional array derived[][][][] as a memo table.
 */

int derive(int table[][MaxInputLength+1][MaxInputLength+1], int i, int j, int k, int l)
{
  int m, n, c;
  int charCount[AlphabetSize];
  int charCount1[AlphabetSize];
  int charCount2[AlphabetSize];

  /* check the memo table to see if the answer is already known */
  if (derived[i][j][k][l] == 1) return 1;
  if (derived[i][j][k][l] == 0) return 0;

  for (c = 0; c < AlphabetSize; c++) {
    charCount[c] = table[c][i][j] + table[c][k][l];
  }
  /* the following check is actually not necessary */
  if (exceedsNorm(charCount)) return 0;

  /* terminating rule */
  /* check to see if w[i,j]w[k,l] is short */
  if (member(0, charCount, AlphabetSize)) {
    derived[i][j][k][l] = 1;
    printf("%d %d %d %d\n", i, j, k, l);
    return 1;
  }

  /* wrapping */
  for (m = i; m <= j; m++) {
    for (n = k; n <= l; n++) {
      if ((i == m && n == l) || (m == j && k == n)) continue;
      for (c = 0; c < AlphabetSize; c++) {
	charCount1[c] = table[c][i][m] + table[c][n][l];
	charCount2[c] = table[c][m][j] + table[c][k][n];
      }
      if (exceedsNorm(charCount1) || exceedsNorm(charCount2)) continue;
      if (derive(table, i, m, n, l) == 1 && derive(table, m, j, k, n) == 1) {
	derived[i][j][k][l] = 1;
	printf("%d %d %d %d\n", i, j, k, l);
	return 1;
      }
    }
  }

  /* left concatenation */
  for (m = i; m <= j; m++) {
    for (n = m; n <= j; n++) {
      if ((i == m && m == n) || (n == j && k == l)) continue;
      for (c = 0; c < AlphabetSize; c++) {
	charCount1[c] = table[c][i][m] + table[c][m][n];
	charCount2[c] = table[c][n][j] + table[c][k][l];
      }
      if (exceedsNorm(charCount1) || exceedsNorm(charCount2)) continue;
      if (derive(table, i, m, m, n) == 1 && derive(table, n, j, k, l) == 1) {
	derived[i][j][k][l] = 1;
	printf("%d %d %d %d\n", i, j, k, l);
	return 1;
      }
    }
  }

  /* right concatenation */
  for (m = k; m <= l; m++) {
    for (n = m; n <= l; n++) {
      if ((i == j && k == m) || (m == n && n == l)) continue;
      for (c = 0; c < AlphabetSize; c++) {
	charCount1[c] = table[c][i][j] + table[c][k][m];
	charCount2[c] = table[c][m][n] + table[c][n][l];
      }
      if (exceedsNorm(charCount1) || exceedsNorm(charCount2)) continue;
      if (derive(table, i, j, k, m) == 1 && derive(table, m, n, n, l) == 1) {
	derived[i][j][k][l] = 1;
	printf("%d %d %d %d\n", i, j, k, l);
	return 1;
      }
    }
  }

  derived[i][j][k][l] = 0;
  printf("%d %d %d %d *\n", i, j, k, l);
  return 0;
}

/*
 * Function: exceedsNorm
 * ---------------------
 * Returns 1 if the infinity norm of the vector (a_count - c_count, b_count - c_count) exceeds
 * the specified limit.
 */

int exceedsNorm(int counts[])
{
  return (abs(counts['a' - FirstCharacter] - counts['c' - FirstCharacter]) > Norm ||
	  abs(counts['b' - FirstCharacter] - counts['c' - FirstCharacter]) > Norm);
}

/*
 * Function: member(int n, int list[], int l)
 * ------------------------------------------
 * Returns 1 if n is one of the first l members of list.
 */

int member (int n, int list[], int l)
{
  int i;
  
  for (i = 0; i < l; i++) {
    if (n == list[i]) return 1;
  }

  return 0;
}

/*
 * Function: isInMIX
 * -----------------
 * Returns 1 if the string represented by the input table is in MIX.
 */

int isInMIX(int table[][MaxInputLength+1][MaxInputLength+1], int inputLength)
{
  return (table['a' - FirstCharacter][0][inputLength] == table['c' - FirstCharacter][0][inputLength] &&
	  table['b' - FirstCharacter][0][inputLength] == table['c' - FirstCharacter][0][inputLength]);
}
